//////////////////////////////////////////////////////////////////////////////////////////////////
// LCD_simple.c - plik pokazujcy prost obsug wywietlacza alfanumerycznego LCD
// 
// Autor: Radosaw Koppel          Kompilator: WinAVR 20050214
//////////////////////////////////////////////////////////////////////////////////////////////////
// Komentarz dodatkowy:
//  Pierwsza wersja. Zaoenie, e wyprowadzenia portu podpito "po koleji"
//  Brak wygodnych komend, wszystko realizowane na do niskim poziomie. 
//////////////////////////////////////////////////////////////////////////////////////////////////
#include <avr\io.h>
#include <inttypes.h>

// Makra upraszczajce dostp do portw
// *** Port
#define PORT(x) XPORT(x)
#define XPORT(x) (PORT##x)
// *** Pin
#define PIN(x) XPIN(x)
#define XPIN(x) (PIN##x)
// *** DDR
#define DDR(x) XDDR(x)
#define XDDR(x) (DDR##x)

// Definicje wyprowadze
#define LCD_RS 2
#define LCD_RSPORT B
#define LCD_E 1
#define LCD_EPORT B
// Port danych: zakadam, e wyprowadzenia s poczone kolejno. 
// Definiowane jest jedynie przesunicie oznaczajce na ktrym wyprowadzeniu jest D4
#define LCD_DPORT B
#define LCD_D4 4 // D5, D6, D7 - poczone kolejno

// Komendy sterujce wywietlaczem
#define LCDC_CLS		0x01
#define LCDC_HOME		0x02
#define LCDC_MODE		0x04
	#define LCDC_MODER		0x02
	#define LCDC_MODEL		0
	#define LCDC_MODEMOVE	0x01
#define LCDC_ON 		0x08
	#define LCDC_ONDISPLAY	0x04
	#define LCDC_ONCURSOR	0x02
	#define LCDC_ONBLINK	0x01
#define LCDC_SHIFT		0x10
	#define LCDC_SHIFTDISP	0x08
	#define LCDC_SHIFTR		0x04
	#define LCDC_SHIFTL		0
#define LCDC_FUNC		0x20
	#define LCDC_FUNC8b		0x10
	#define LCDC_FUNC4b		0
	#define LCDC_FUNC2L		0x08
	#define LCDC_FUNC1L		0
	#define LCDC_FUNC5x10	0x4
	#define LCDC_FUNC5x7	0
#define LCDC_CGA		0x40
#define LCDC_DDA		0x80

//--------------------------------------------------
// Generowanie opnie
#define delay250ns() {asm volatile("nop"::);}

#define delayus8(t)\
	{asm volatile( \
		"delayus8_loop%=: \n\t"\
			"nop \n\t"\
			"dec %[ticks] \n\t"\
			"brne delayus8_loop%= \n\t"\
	: :[ticks]"r"(t) );}
	// DEC - 1 cykl, BRNE 2 cykle, + 1xnop. Zegar 4MHz

void delay100us8(uint8_t t)
{
	while(t>0)
	{
		delayus8(100); 
		--t; 
	}
}
// Koniec opnie
//--------------------------------------------------

//--------------------------------------------------
// Obsuga wywietlacza
// Funkcje niskiego poziomu
#define LCD_EPULSE() \
	{PORT(LCD_EPORT) |= 1<<LCD_E; \
	delay250ns(); \
	PORT(LCD_EPORT) &= ~(1<<LCD_E);}

void LCDsendHalf(uint8_t data)
{
	data = (data & 0x0F) << LCD_D4; 
	//data |= 0xF0;
	//data &= 0x0F;
	//data <<= LCD_D4; 
	PORT(LCD_DPORT) = 
		(PORT(LCD_DPORT) & ~(0x0F<<LCD_D4)) | data;
	//PORT(LCD_DPORT) &= ~(0x0F<<LCD_D4); 
	//PORT(LCD_DPORT) |= data; 
	LCD_EPULSE();
}

void LCDsend(uint8_t data)
{
	// Starsza cz
	LCDsendHalf(data>>4);
	// Modsza cz
	LCDsendHalf(data);
	delayus8(120);
}

// Funkcje interfejsu
void LCDcommand(uint8_t command)
{
	PORT(LCD_RSPORT) &= ~(1<<LCD_RS);
	LCDsend(command);
}

void LCDdata(uint8_t data)
{
	PORT(LCD_RSPORT) |= 1<<LCD_RS;
	LCDsend(data);
}

void LCDcls(void)
{
	LCDcommand(LCDC_CLS);
	delay100us8(48);
}

void LCDhome(void)
{
	LCDcommand(LCDC_HOME);
	delay100us8(48);
}

void LCDinit(void)
{
	delay100us8(150);
	PORT(LCD_RSPORT) &= ~(1<<LCD_RS);
	LCDsendHalf(LCDC_FUNC|LCDC_FUNC8b); 
	delay100us8(41);
	LCDsendHalf(LCDC_FUNC|LCDC_FUNC8b);
	delay100us8(2);
	LCDsendHalf(LCDC_FUNC|LCDC_FUNC4b);
	delay100us8(2);
	// Teraz jest ju 4b. Koniec korzystania z sendHalf
	LCDcommand(LCDC_FUNC|LCDC_FUNC4b|LCDC_FUNC2L|LCDC_FUNC5x7);
	LCDcommand(LCDC_ON);
	LCDcls();
	LCDcommand(LCDC_MODE|LCDC_MODER);
	LCDcommand(LCDC_ON|LCDC_ONDISPLAY|LCDC_ONCURSOR);
}

// Koniec obsugi wywietlacza
//--------------------------------------------------

// START
int main(void)
{
	//----------------------------------------------
	// Inicjacja - uwaga: uproszczona, moe wymaga 
	// zmian przy zmianie przyporzdkowania portw
	DDR(LCD_DPORT) = 1<<LCD_E | 1<<LCD_RS | 
		0x0F<<LCD_D4;
	
	LCDinit();
	// Koniec inicjacji
	//----------------------------------------------
	
	// \/ Bez tego, czasami zerowanie z poziomu programatora
	// \/ powoduje mieci na wywietlaczu. 
	LCDcls(); 
	
	LCDdata('H');
	LCDdata('E');
	LCDdata('L');
	LCDdata('L');
	LCDdata('O');
	LCDdata(' ');
	LCDdata(':');
	LCDdata('-');
	LCDdata(')');
	return 0;
}
